# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Helper functions for writing hooks in python."""

__metaclass__ = type
__all__ = [
    'apt_get_install',
    'command',
    'DictDiffer',
    'generate_string',
    'get_config',
    'get_value_from_line',
    'grep',
    'install_extra_repository',
    'log',
    'run',
    'relation_get',
    'relation_set',
    'unit_info',
    ]

import json
import os
import subprocess
import sys
from textwrap import dedent
import yaml


def command(*base_args):
    def callable_command(*args):
        all_args = base_args + args
        return subprocess.check_output(all_args, shell=False)

    return callable_command


log = command('juju-log')
apt_get_install = command('apt-get', 'install', '-y', '--force-yes')


def run(*args):
    log('run: ' + str(args))
    return subprocess.check_output(args, shell=False)


def get_config():
    config_get = command('config-get', '--format=json')
    return json.loads(config_get())


def install_extra_repository(extra_repository):
    try:
        run('apt-add-repository', extra_repository)
        run('apt-get', 'update')
    except subprocess.CalledProcessError as e:
        log("Error adding repository %s" % extra_repository)
        log(e)
        sys.exit(-1)


def relation_get(*args):
    cmd = command('relation-get')
    return cmd(*args).strip()


def relation_set(**kwargs):
    cmd = command('relation-set')
    args = ['{}={}'.format(k, v) for k, v in kwargs.items()]
    return cmd(*args)


def unit_info(service_name, item_name, data=None):
    if data is None:
        data = yaml.safe_load(run('juju', 'status'))
    services = data['services'][service_name]
    units = services['units']
    item = units.items()[0][1][item_name]
    return item


def grep(content, filename):
    with open(filename) as f:
        for line in f:
            if re.match(content, line):
                return line.strip()


def get_value_from_line(line):
    return line.split('=')[1].strip('"\' ')


class DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    # Based on answer by hughdbrown at:
    # http://stackoverflow.com/questions/1165352

    def __init__(self, current_dict, past_dict):
        self.current_dict = current_dict
        self.past_dict = past_dict
        self.set_current = set(current_dict)
        self.set_past = set(past_dict)
        self.intersect = self.set_current.intersection(self.set_past)

    @property
    def added(self):
        return self.set_current - self.intersect

    @property
    def removed(self):
        return self.set_past - self.intersect

    @property
    def changed(self):
        return set(key for key in self.intersect
                   if self.past_dict[key] != self.current_dict[key])
    @property
    def unchanged(self):
        return set(key for key in self.intersect
                   if self.past_dict[key] == self.current_dict[key])
    @property
    def modified(self):
        return self.current_dict != self.past_dict

    @property
    def added_or_changed(self):
        return self.added.union(self.changed)

    def _changes(self, keys):
        new = {}
        old = {}
        for k in keys:
            new[k] = self.current_dict.get(k)
            old[k] = self.past_dict.get(k)
        return "%s -> %s" % (old, new)
    def __str__(self):
        if self.modified:
            s = dedent("""\
            added: %s
            removed: %s
            changed: %s
            unchanged: %s""") % (
                self._changes(self.added),
                self._changes(self.removed),
                self._changes(self.changed),
                list(self.unchanged))
        else:
            s = "no changes"
        return s


class Serializer:

    def __init__(self, path, default=None, serialize=None, deserialize=None):
        self.path = path
        log("Serializer path: " + self.path)
        self.default = default or {}
        self.serialize = serialize or json.dump
        self.deserialize = deserialize or json.load

    def exists(self):
        return os.path.exists(self.path)

    def get(self):
        if self.exists():
            with open(self.path) as f:
                return self.deserialize(f)
        return self.default

    def set(self, data):
        with open(self.path, 'w') as f:
            self.serialize(data, f)
